@@ -27,14 +27,15 @@
 /*
 
 Icons are just like BMP's, except for the header.
- 
+
 Known bugs:
-	* bi-tonal files aren't tested 
+	* bi-tonal files aren't tested
 
 */
 
 #include <config.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
 #include "gdk-pixbuf-private.h"
@@ -42,10 +43,10 @@ Known bugs:
 
 
 
-/* 
+/*
 
 These structures are actually dummies. These are according to
-the "Windows API reference guide volume II" as written by 
+the "Windows API reference guide volume II" as written by
 Borland International, but GCC fiddles with the alignment of
 the internal members.
 
@@ -73,14 +74,15 @@ struct BitmapInfoHeader {
 };
 
 #ifdef DUMPBIH
-/* 
+/*
 
-DumpBIH printf's the values in a BitmapInfoHeader to the screen, for 
+DumpBIH printf's the values in a BitmapInfoHeader to the screen, for
 debugging purposes.
 
 */
-static void DumpBIH(unsigned char *BIH)
-{				
+static void
+DumpBIH (unsigned char *BIH)
+{
 	printf("biSize      = %i \n",
 	       (int)(BIH[3] << 24) + (BIH[2] << 16) + (BIH[1] << 8) + (BIH[0]));
 	printf("biWidth     = %i \n",
@@ -116,7 +118,7 @@ struct headerpair {
 	guint width;
 	guint height;
 	guint depth;
-	guint Negative;		/* Negative = 1 -> top down BMP,  
+	guint Negative;		/* Negative = 1 -> top down BMP,
 				   Negative = 0 -> bottom up BMP */
 };
 
@@ -135,15 +137,18 @@ struct ico_progressive_state {
 	gint LineDone;		/* # of bytes in LineBuf */
 	gint Lines;		/* # of finished lines */
 
-	gint Type;		/*  
+	gint Type;		/*
+				   32 = RGB + alpha
 				   24 = RGB
+				   16 = 555 RGB
 				   8 = 8 bit colormapped
-				   1  = 1 bit bitonal 
+				   4  = 4 bpp colormapped
+				   1  = 1 bit bitonal
 				 */
 
 
 	struct headerpair Header;	/* Decoded (BE->CPU) header */
-	
+
 	gint			DIBoffset;
 	gint			ImageScore;
 
@@ -151,40 +156,55 @@ struct ico_progressive_state {
 	GdkPixbuf *pixbuf;	/* Our "target" */
 };
 
-gpointer
-gdk_pixbuf__ico_image_begin_load(ModulePreparedNotifyFunc prepared_func,
-				 ModuleUpdatedNotifyFunc updated_func,
-				 ModuleFrameDoneNotifyFunc frame_done_func,
-				 ModuleAnimationDoneNotifyFunc anim_done_func,
-				 gpointer user_data);
+gpointer gdk_pixbuf__ico_image_begin_load (ModulePreparedNotifyFunc prepared_func,
+					   ModuleUpdatedNotifyFunc updated_func,
+					   ModuleFrameDoneNotifyFunc frame_done_func,
+					   ModuleAnimationDoneNotifyFunc anim_done_func,
+					   gpointer user_data);
 void gdk_pixbuf__ico_image_stop_load(gpointer data);
 gboolean gdk_pixbuf__ico_image_load_increment(gpointer data, guchar * buf, guint size);
 
 
+static void 
+context_free (struct ico_progressive_state *context)
+{
+	if (context->LineBuf != NULL)
+		free (context->LineBuf);
+	context->LineBuf = NULL;
+	if (context->HeaderBuf != NULL)
+		free (context->HeaderBuf);
+
+	if (context->pixbuf)
+		gdk_pixbuf_unref (context->pixbuf);
+
+	g_free (context);
+}
 
 /* Shared library entry point --> Can go when generic_image_load
    enters gdk-pixbuf-io */
 GdkPixbuf *
-gdk_pixbuf__ico_image_load(FILE * f)
+gdk_pixbuf__ico_image_load (FILE *f)
 {
-	guchar *membuf;
+	guchar membuf[4096];
 	size_t length;
 	struct ico_progressive_state *State;
 
 	GdkPixbuf *pb;
 
 	State = gdk_pixbuf__ico_image_begin_load(NULL, NULL, NULL, NULL, NULL);
-	membuf = g_malloc(4096);
 
-	g_assert(membuf != NULL);
+	if (State == NULL)
+		return NULL;
 
 	while (feof(f) == 0) {
-		length = fread(membuf, 1, 4096, f);
+		length = fread(membuf, 1, sizeof (membuf), f);
 		if (length > 0)
-			gdk_pixbuf__ico_image_load_increment(State, membuf, length);
-
+			if (!gdk_pixbuf__ico_image_load_increment(State, membuf, length)) {
+				gdk_pixbuf__ico_image_stop_load (State);
+				return NULL;
+			}
 	}
-	g_free(membuf);
+
 	if (State->pixbuf != NULL)
 		gdk_pixbuf_ref(State->pixbuf);
 
@@ -194,95 +214,108 @@ gdk_pixbuf__ico_image_load(FILE * f)
 	return pb;
 }
 
-static void DecodeHeader(guchar *Data, gint Bytes,
-			 struct ico_progressive_state *State)
+static gboolean
+DecodeHeader (guchar *Data, gint Bytes, struct ico_progressive_state *State)
 {
 /* For ICO's we have to be very clever. There are multiple images possible
    in an .ICO. For now, we select (in order of priority):
      1) The one with the highest number of colors
      2) The largest one
- */   
- 
+ */
+
 	gint IconCount = 0; /* The number of icon-versions in the file */
 	guchar *BIH; /* The DIB for the used icon */
  	guchar *Ptr;
  	gint I;
- 
+
  	/* Step 1: The ICO header */
- 
+
  	IconCount = (Data[5] << 8) + (Data[4]);
- 
+
  	State->HeaderSize = 6 + IconCount*16;
- 	
+
  	if (State->HeaderSize>State->BytesInHeaderBuf) {
- 		State->HeaderBuf=g_realloc(State->HeaderBuf,State->HeaderSize);
+		guchar *tmp=realloc(State->HeaderBuf,State->HeaderSize);
+		if (!tmp)
+			return FALSE;
+		State->HeaderBuf = tmp;
  		State->BytesInHeaderBuf = State->HeaderSize;
  	}
  	if (Bytes < State->HeaderSize)
- 		return;
- 	
- 	/* We now have all the "short-specs" of the versions 
+ 		return TRUE;
+
+ 	/* We now have all the "short-specs" of the versions
  	   So we iterate through them and select the best one */
- 	   
+
  	State->ImageScore = 0;
  	State->DIBoffset  = 0;
  	Ptr = Data + 6;
 	for (I=0;I<IconCount;I++) {
 		int ThisWidth, ThisHeight,ThisColors;
 		int ThisScore;
-		
+
 		ThisWidth = Ptr[0];
 		ThisHeight = Ptr[1];
 		ThisColors = (Ptr[2]);
-		if (ThisColors==0) 
+		if (ThisColors==0)
 			ThisColors=256; /* Yes, this is in the spec, ugh */
-		
-		ThisScore = ThisColors*1024+ThisWidth*ThisHeight; 
+
+		ThisScore = ThisColors*1024+ThisWidth*ThisHeight;
 
 		if (ThisScore>State->ImageScore) {
 			State->ImageScore = ThisScore;
 			State->DIBoffset = (Ptr[15]<<24)+(Ptr[14]<<16)+
 					   (Ptr[13]<<8) + (Ptr[12]);
-								 
+
 		}
-		
-		
-		Ptr += 16;	
-	} 
-		
+
+
+		Ptr += 16;
+	}
+
+	if (State->DIBoffset < 0)
+		return FALSE; /* Invalid header */
+
 	/* We now have a winner, pointed to in State->DIBoffset,
 	   so we know how many bytes are in the "header" part. */
-	      
+
 	State->HeaderSize = State->DIBoffset + 40; /* 40 = sizeof(InfoHeader) */
-	
+
  	if (State->HeaderSize>State->BytesInHeaderBuf) {
- 		State->HeaderBuf=g_realloc(State->HeaderBuf,State->HeaderSize);
+		guchar *tmp=realloc(State->HeaderBuf,State->HeaderSize);
+		if (!tmp)
+			return FALSE;
+		State->HeaderBuf = tmp;
  		State->BytesInHeaderBuf = State->HeaderSize;
  	}
-	if (Bytes<State->HeaderSize) 
-		return;   
-	
+	if (Bytes<State->HeaderSize)
+		return TRUE;
+
 	BIH = Data+State->DIBoffset;
 
 #ifdef DUMPBIH
 	DumpBIH(BIH);
-#endif	
-	
+#endif
+
 	/* Add the palette to the headersize */
-		
+
 	State->Header.width =
 	    (int)(BIH[7] << 24) + (BIH[6] << 16) + (BIH[5] << 8) + (BIH[4]);
+	if (State->Header.width == 0)
+		return FALSE;
+
 	State->Header.height =
 	    (int)(BIH[11] << 24) + (BIH[10] << 16) + (BIH[9] << 8) + (BIH[8])/2;
 	    /* /2 because the BIH height includes the transparency mask */
+	if (State->Header.height == 0)
+		return FALSE;
+
 	State->Header.depth = (BIH[15] << 8) + (BIH[14]);;
 
-	State->Type = State->Header.depth;	
+	State->Type = State->Header.depth;
 	if (State->Lines>=State->Header.height)
 		State->Type = 1; /* The transparency mask is 1 bpp */
-	
-	
-	
+
 	/* Determine the  palette size. If the header indicates 0, it
 	   is actually the maximum for the bpp. You have to love the
 	   guys who made the spec. */
@@ -294,19 +327,23 @@ static void DecodeHeader(guchar *Data, gint Bytes,
 		I = 16*4;
 	if ((I==0)&&(State->Type==8))
 		I = 256*4;
-	
+
 	State->HeaderSize+=I;
-	
+
  	if (State->HeaderSize>State->BytesInHeaderBuf) {
- 		State->HeaderBuf=g_realloc(State->HeaderBuf,State->HeaderSize);
+		guchar *tmp=realloc(State->HeaderBuf,State->HeaderSize);
+		if (!tmp)
+			return FALSE;
+		State->HeaderBuf = tmp;
  		State->BytesInHeaderBuf = State->HeaderSize;
  	}
  	if (Bytes < State->HeaderSize)
- 		return;
+ 		return TRUE;
 
 	if ((BIH[16] != 0) || (BIH[17] != 0) || (BIH[18] != 0)
 	    || (BIH[19] != 0)) {
-		g_assert(0); /* Compressed icons aren't allowed */
+		/* Compressed icons are not supported */
+		return FALSE;
 	}
 
 	/* Negative heights mean top-down pixel-order */
@@ -318,17 +355,23 @@ static void DecodeHeader(guchar *Data, gint Bytes,
 		State->Header.width = -State->Header.width;
 	}
 
-	if (State->Type == 24)
+	if (State->Type == 32)
+		State->LineWidth = State->Header.width * 4;
+	else if (State->Type == 24)
 		State->LineWidth = State->Header.width * 3;
-	if (State->Type == 8)
+	else if (State->Type == 16)
+		State->LineWidth = State->Header.height * 2;
+	else if (State->Type == 8)
 		State->LineWidth = State->Header.width * 1;
-	if (State->Type == 4) {
+	else if (State->Type == 4)
 		State->LineWidth = (State->Header.width+1)/2;
-	}
-	if (State->Type == 1) {
+	else if (State->Type == 1) {
 		State->LineWidth = State->Header.width / 8;
 		if ((State->Header.width & 7) != 0)
 			State->LineWidth++;
+	} else {
+		/* Unsupported ICO type */
+		return FALSE;
 	}
 
 	/* Pad to a 32 bit boundary */
@@ -336,17 +379,21 @@ static void DecodeHeader(guchar *Data, gint Bytes,
 		State->LineWidth = (State->LineWidth / 4) * 4 + 4;
 
 
-	if (State->LineBuf == NULL)
-		State->LineBuf = g_malloc(State->LineWidth);
+	if (State->LineBuf == NULL) {
+		State->LineBuf = malloc(State->LineWidth);
+		if (!State->LineBuf)
+			return FALSE;
+	}
 
 	g_assert(State->LineBuf != NULL);
 
 
 	if (State->pixbuf == NULL) {
-		State->pixbuf =
-		    gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8,
-				   (gint) State->Header.width,
-				   (gint) State->Header.height);
+		State->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
+						(gint) State->Header.width,
+						(gint) State->Header.height);
+		if (!State->pixbuf)
+			return FALSE;
 
 		if (State->prepared_func != NULL)
 			/* Notify the client that we are ready to go */
@@ -355,20 +402,21 @@ static void DecodeHeader(guchar *Data, gint Bytes,
 
 	}
 
+	return TRUE;
 }
 
-/* 
+/*
  * func - called when we have pixmap created (but no image data)
  * user_data - passed as arg 1 to func
  * return context (opaque to user)
  */
 
 gpointer
-gdk_pixbuf__ico_image_begin_load(ModulePreparedNotifyFunc prepared_func,
-				 ModuleUpdatedNotifyFunc updated_func,
-				 ModuleFrameDoneNotifyFunc frame_done_func,
-				 ModuleAnimationDoneNotifyFunc anim_done_func,
-				 gpointer user_data)
+gdk_pixbuf__ico_image_begin_load (ModulePreparedNotifyFunc prepared_func,
+				  ModuleUpdatedNotifyFunc updated_func,
+				  ModuleFrameDoneNotifyFunc frame_done_func,
+				  ModuleAnimationDoneNotifyFunc anim_done_func,
+				  gpointer user_data)
 {
 	struct ico_progressive_state *context;
 
@@ -378,7 +426,11 @@ gdk_pixbuf__ico_image_begin_load(ModulePreparedNotifyFunc prepared_func,
 	context->user_data = user_data;
 
 	context->HeaderSize = 54;
-	context->HeaderBuf = g_malloc(14 + 40 + 4*256 + 512);
+	context->HeaderBuf = malloc(14 + 40 + 4*256 + 512);
+	if (!context->HeaderBuf) {
+		g_free (context);
+		return NULL;
+	}
 	/* 4*256 for the colormap */
 	context->BytesInHeaderBuf = 14 + 40 + 4*256 + 512 ;
 	context->HeaderDone = 0;
@@ -404,28 +456,45 @@ gdk_pixbuf__ico_image_begin_load(ModulePreparedNotifyFunc prepared_func,
  *
  * free context, unref gdk_pixbuf
  */
-void gdk_pixbuf__ico_image_stop_load(gpointer data)
+void
+gdk_pixbuf__ico_image_stop_load (gpointer data)
 {
 	struct ico_progressive_state *context =
 	    (struct ico_progressive_state *) data;
 
-
 	g_return_if_fail(context != NULL);
 
-	if (context->LineBuf != NULL)
-		g_free(context->LineBuf);
-	context->LineBuf = NULL;
-	if (context->HeaderBuf != NULL)
-		g_free(context->HeaderBuf);
-
-	if (context->pixbuf)
-		gdk_pixbuf_unref(context->pixbuf);
-
-	g_free(context);
+	context_free (context);
 }
 
 
-static void OneLine24(struct ico_progressive_state *context)
+static void
+OneLine32 (struct ico_progressive_state *context)
+{
+	gint X;
+	guchar *Pixels;
+
+	X = 0;
+	if (context->Header.Negative == 0)
+		Pixels = (context->pixbuf->pixels +
+			  context->pixbuf->rowstride *
+			  (context->Header.height - context->Lines - 1));
+	else
+		Pixels = (context->pixbuf->pixels +
+			  context->pixbuf->rowstride *
+			  context->Lines);
+	while (X < context->Header.width) {
+		Pixels[X * 4 + 0] = context->LineBuf[X * 4 + 2];
+		Pixels[X * 4 + 1] = context->LineBuf[X * 4 + 1];
+		Pixels[X * 4 + 2] = context->LineBuf[X * 4 + 0];
+		Pixels[X * 4 + 3] = context->LineBuf[X * 4 + 3];
+		X++;
+	}
+
+}
+
+static void
+OneLine24 (struct ico_progressive_state *context)
 {
 	gint X;
 	guchar *Pixels;
@@ -448,7 +517,45 @@ static void OneLine24(struct ico_progressive_state *context)
 
 }
 
-static void OneLine8(struct ico_progressive_state *context)
+static void
+OneLine16 (struct ico_progressive_state *context)
+{
+	int i;
+	guchar *pixels;
+	guchar *src;
+
+	if (context->Header.Negative == 0)
+		pixels = (context->pixbuf->pixels +
+			  context->pixbuf->rowstride * (context->Header.height - context->Lines - 1));
+	else
+		pixels = (context->pixbuf->pixels +
+			  context->pixbuf->rowstride * context->Lines);
+
+	src = context->LineBuf;
+
+	for (i = 0; i < context->Header.width; i++) {
+		int v, r, g, b;
+
+		v = (int) src[0] | ((int) src[1] << 8);
+		src += 2;
+
+		/* Extract 5-bit RGB values */
+
+		r = (v >> 10) & 0x1f;
+		g = (v >> 5) & 0x1f;
+		b = v & 0x1f;
+
+		/* Fill the rightmost bits to form 8-bit values */
+
+		*pixels++ = (r << 3) | (r >> 2);
+		*pixels++ = (g << 3) | (g >> 2);
+		*pixels++ = (b << 3) | (b >> 2);
+		pixels++; /* skip alpha channel */
+	}
+}
+
+static void
+OneLine8 (struct ico_progressive_state *context)
 {
 	gint X;
 	guchar *Pixels;
@@ -473,7 +580,9 @@ static void OneLine8(struct ico_progressive_state *context)
 		X++;
 	}
 }
-static void OneLine4(struct ico_progressive_state *context)
+
+static void
+OneLine4 (struct ico_progressive_state *context)
 {
 	gint X;
 	guchar *Pixels;
@@ -487,10 +596,10 @@ static void OneLine4(struct ico_progressive_state *context)
 		Pixels = (context->pixbuf->pixels +
 			  context->pixbuf->rowstride *
 			  context->Lines);
-	
+
 	while (X < context->Header.width) {
 		guchar Pix;
-		
+
 		Pix = context->LineBuf[X/2];
 
 		Pixels[X * 4 + 0] =
@@ -500,7 +609,7 @@ static void OneLine4(struct ico_progressive_state *context)
 		Pixels[X * 4 + 2] =
 		    context->HeaderBuf[4 * (Pix>>4) + 40+context->DIBoffset];
 		X++;
-		if (X<context->Header.width) { 
+		if (X<context->Header.width) {
 			/* Handle the other 4 bit pixel only when there is one */
 			Pixels[X * 4 + 0] =
 			    context->HeaderBuf[4 * (Pix&15) + 42+context->DIBoffset];
@@ -511,10 +620,11 @@ static void OneLine4(struct ico_progressive_state *context)
 			X++;
 		}
 	}
-	
+
 }
 
-static void OneLine1(struct ico_progressive_state *context)
+static void
+OneLine1 (struct ico_progressive_state *context)
 {
 	gint X;
 	guchar *Pixels;
@@ -541,7 +651,8 @@ static void OneLine1(struct ico_progressive_state *context)
 	}
 }
 
-static void OneLineTransp(struct ico_progressive_state *context)
+static void
+OneLineTransp (struct ico_progressive_state *context)
 {
 	gint X;
 	guchar *Pixels;
@@ -562,7 +673,7 @@ static void OneLineTransp(struct ico_progressive_state *context)
 		Bit = Bit & 1;
 		/* The joys of having a BGR byteorder */
 		Pixels[X * 4 + 3] = 255-Bit*255;
-#if 0		
+#if 0
 		if (Bit){
 		  Pixels[X*4+0] = 255;
 		  Pixels[X*4+1] = 255;
@@ -570,35 +681,39 @@ static void OneLineTransp(struct ico_progressive_state *context)
 		  Pixels[X*4+0] = 0;
 		  Pixels[X*4+1] = 0;
 		}
-#endif		
+#endif
 		X++;
 	}
 }
 
 
-static void OneLine(struct ico_progressive_state *context)
+static void
+OneLine (struct ico_progressive_state *context)
 {
 	context->LineDone = 0;
-	
+
 	if (context->Lines >= context->Header.height*2) {
 		return;
 	}
-		
-	if (context->Lines <context->Header.height) {		
 
-		if (context->Type == 24)
-			OneLine24(context);
-		if (context->Type == 8)
-			OneLine8(context);
-		if (context->Type == 4)
-			OneLine4(context);
-		if (context->Type == 1)
-			OneLine1(context);
+	if (context->Lines < context->Header.height) {
+		if (context->Type == 32)
+			OneLine32 (context);
+		else if (context->Type == 24)
+			OneLine24 (context);
+		else if (context->Type == 16)
+			OneLine16 (context);
+		else if (context->Type == 8)
+			OneLine8 (context);
+		else if (context->Type == 4)
+			OneLine4 (context);
+		else if (context->Type == 1)
+			OneLine1 (context);
+		else
+			g_assert_not_reached ();
 	} else
-	{
 		OneLineTransp(context);
-	}
-	
+
 	context->Lines++;
 	if (context->Lines>=context->Header.height) {
 	 	context->Type = 1;
@@ -608,9 +723,9 @@ static void OneLine(struct ico_progressive_state *context)
 		/* Pad to a 32 bit boundary */
 		if (((context->LineWidth % 4) > 0))
 			context->LineWidth = (context->LineWidth / 4) * 4 + 4;
-			
+
 	}
-	  
+
 
 	if (context->updated_func != NULL) {
 		(*context->updated_func) (context->pixbuf,
@@ -631,7 +746,7 @@ static void OneLine(struct ico_progressive_state *context)
  * append image data onto inrecrementally built output image
  */
 gboolean
-gdk_pixbuf__ico_image_load_increment(gpointer data, guchar * buf, guint size)
+gdk_pixbuf__ico_image_load_increment (gpointer data, guchar * buf, guint size)
 {
 	struct ico_progressive_state *context =
 	    (struct ico_progressive_state *) data;
@@ -640,7 +755,7 @@ gdk_pixbuf__ico_image_load_increment(gpointer data, guchar * buf, guint size)
 
 	while (size > 0) {
 		g_assert(context->LineDone >= 0);
-		if (context->HeaderDone < context->HeaderSize) {	/* We still 
+		if (context->HeaderDone < context->HeaderSize) {	/* We still
 									   have headerbytes to do */
 			BytesToCopy =
 			    context->HeaderSize - context->HeaderDone;
@@ -655,7 +770,7 @@ gdk_pixbuf__ico_image_load_increment(gpointer data, guchar * buf, guint size)
 			context->HeaderDone += BytesToCopy;
 
 		}  else
-		{  
+		{
 			BytesToCopy =
 			    context->LineWidth - context->LineDone;
 			if (BytesToCopy > size)
@@ -677,11 +792,10 @@ gdk_pixbuf__ico_image_load_increment(gpointer data, guchar * buf, guint size)
 
 		}
 
-		if (context->HeaderDone >= 6)
-			DecodeHeader(context->HeaderBuf,
-				     context->HeaderDone, context);
-
-
+		if (context->HeaderDone >= 6) {
+			if (!DecodeHeader(context->HeaderBuf, context->HeaderDone, context))
+				return FALSE;
+		}
 	}
 
 	return TRUE;
